@@ -683,7 +683,7 @@ class DataStream:
             left_on (str): the name of the join column in this DataStream.
             right_on (str): the name of the join column in `right`.
             suffix (str): if `right` has columns with the same names as columns in this DataStream, their names will be appended with the suffix in the result.
-            how (str): only supports "inner" for now. 
+            how (str): supports "inner", "left", "semi" or "anti"
 
         Return:
             A new DataStream that's the joined result of this DataStream and "right". By default, columns from both side will be retained, 
@@ -702,6 +702,7 @@ class DataStream:
             ~~~
         """
 
+        assert how in {"inner", "left", "semi", "anti"}
         assert type(right) == polars.internals.DataFrame or issubclass(
             type(right), DataStream)
         
@@ -745,8 +746,15 @@ class DataStream:
             else:
                 new_schema.append(col)
                 schema_mapping[col] = (right_table_id, col)
+        
+        # you only need the key column on the RHS! select overloads in DataStream or Polars DataFrame runtime polymorphic
+        if how == "semi" or how == "anti":
+            right = right.select([right_on])
 
         if issubclass(type(right), DataStream):
+
+            operator = JoinExecutor(on, left_on, right_on, suffix=suffix, how=how) if how != "anti" else AntiJoinExecutor(on , left_on, right_on, suffix = suffix)
+
             return self.quokka_context.new_stream(
                 sources={0: self, 1: right},
                 partitioners={0: HashPartitioner(
@@ -755,10 +763,12 @@ class DataStream:
                     schema=new_schema,
                     schema_mapping=schema_mapping,
                     required_columns={0: {left_on}, 1: {right_on}},
-                    operator=PolarJoinExecutor(on, left_on, right_on, suffix=suffix, how=how)),
+                    operator= operator),
                 schema=new_schema,
                 ordering=None)
+
         elif type(right) == polars.internals.DataFrame:
+            
             return self.quokka_context.new_stream(
                 sources={0: self},
                 partitioners={0: PassThroughPartitioner()},
