@@ -14,6 +14,7 @@ import random
 import sys
 from pyarrow.fs import S3FileSystem, LocalFileSystem
 import pyarrow.dataset as ds
+import ray
 
 class Executor:
     def __init__(self) -> None:
@@ -64,7 +65,6 @@ class StorageExecutor(Executor):
     def done(self,executor_id):
         return
 
-
 class OutputExecutor(Executor):
     def __init__(self, filepath, format, prefix = "part", mode = "local", row_group_size = 5000000) -> None:
         self.num = 0
@@ -88,6 +88,7 @@ class OutputExecutor(Executor):
         fs = LocalFileSystem() if self.mode == "local" else S3FileSystem()
         self.my_batches.extend([i for i in batches if i is not None])
 
+
         '''
         You want to use Payrrow's write table API to flush everything at once. to get the parallelism.
         Being able to write multiple row groups at once really speeds things up. It's okay if you are slow at first
@@ -96,6 +97,9 @@ class OutputExecutor(Executor):
 
         lengths = [len(batch) for batch in self.my_batches]
         total_len = np.sum(lengths)
+
+        print(time.time(), len(self.my_batches), total_len)
+
         write_len = total_len // self.row_group_size * self.row_group_size
         cum_sum = np.cumsum(lengths)
         if len(np.where(cum_sum > write_len)[0]) == 0:
@@ -156,7 +160,10 @@ class BroadcastJoinExecutor(Executor):
     def __init__(self, small_table, on = None, small_on = None, big_on = None, suffix = "_small", how = "inner"):
 
         self.suffix = suffix
+
+        assert how in {"inner", "left", "semi"}
         self.how = how
+        self.batch_how = how if how != "left" else "inner"
 
         if type(small_table) == pd.core.frame.DataFrame:
             self.state = polars.from_pandas(small_table)
@@ -164,8 +171,10 @@ class BroadcastJoinExecutor(Executor):
             self.state = small_table
         else:
             raise Exception("small table data type not accepted")
-
-        self.checkpointed = False
+        
+        if how == "left" or how == "anti":
+            self.left_null = None
+            self.first_row_right = small_table[0]
 
         if on is not None:
             assert small_on is None and big_on is None
@@ -177,22 +186,6 @@ class BroadcastJoinExecutor(Executor):
             self.big_on = big_on
         
         assert self.small_on in self.state.columns
-
-
-    def serialize(self):
-        # if you have already checkpointed the small table, don't checkpoint anything
-
-        # otherwise checkpoint the small table
-        if not self.checkpointed:
-            assert self.state is not None
-            self.checkpointed = True
-            return {0:self.state}, "all"
-        else:
-            return None, "inc" # second argument here doesn't really matter
-    
-    def deserialize(self, s):
-        assert type(s) == list
-        self.state = s[0][0]
     
     # the execute function signature does not change. stream_id will be a [0 - (length of InputStreams list - 1)] integer
     def execute(self,batches, stream_id, executor_id):
@@ -201,26 +194,44 @@ class BroadcastJoinExecutor(Executor):
         if len(batches) == 0:
             return
         batch = polars.concat(batches)
-        result = batch.join(self.state, left_on = self.big_on, right_on = self.small_on, how = self.how, suffix = self.suffix)
 
-        if result is not None and len(result) > 0:
+        if self.how != "anti":
+            try:
+                result = batch.join(self.state, left_on = self.big_on, right_on = self.small_on, how = self.batch_how, suffix = self.suffix)
+            except:
+                print(batch, self.state)
+
+        new_left_null = None
+        if self.how == "left" or self.how == "anti":
+            new_left_null = batch.join(self.state1, left_on = self.left_on, right_on= self.right_on, how = "anti", suffix = self.suffix)
+        
+        if (self.how == "left" or self.how == "anti") and new_left_null is not None and len(new_left_null) > 0:
+            if self.left_null is None:
+                self.left_null = new_left_null
+            else:
+                self.left_null.vstack(new_left_null, in_place= True)
+
+        if self.how != "anti" and result is not None and len(result) > 0:
             return result
     
     def done(self,executor_id):
         #print(len(self.state0),len(self.state1))
         #print("done join ", executor_id)
-        pass
+        
+        if (self.how == "left" or self.how == "anti") and self.left_null is not None and len(self.left_null) > 0:
+            if self.how == "left":
+                return self.left_null.join(self.first_row_right, left_on= self.left_on, right_on= self.right_on, how = "left", suffix = self.suffix)
+            if self.how == "anti":
+                return self.left_null
 
 
-class GroupAsOfJoinExecutor():
+class JoinExecutor(Executor):
     # batch func here expects a list of dfs. This is a quark of the fact that join results could be a list of dfs.
     # batch func must return a list of dfs too
-    def __init__(self, group_on= None, group_left_on = None, group_right_on = None, on = None, left_on = None, right_on = None, suffix="_right"):
+    def __init__(self, on = None, left_on = None, right_on = None, suffix="_right", how = "inner"):
 
-        self.trade = {}
-        self.quote = {}
-        self.ckpt_start0 = 0
-        self.ckpt_start1 = 0
+        self.state0 = None
+        self.state1 = None
         self.suffix = suffix
 
         if on is not None:
@@ -232,181 +243,79 @@ class GroupAsOfJoinExecutor():
             self.left_on = left_on
             self.right_on = right_on
         
-        if group_on is not None:
-            assert group_left_on is None and group_right_on is None
-            self.group_left_on = group_on
-            self.group_right_on = group_on
-        else:
-            assert group_left_on is not None and group_right_on is not None
-            self.group_left_on = group_left_on
-            self.group_right_on = group_right_on
+        assert how in {"inner", "left",  "semi"}
+        self.how = how
+        self.batch_how = how if how != "left" else "inner"
+        
+        if how == "left":
+            self.left_null = None
+            self.first_row_right = None # this is a hack to produce the left join NULLs at the end.
 
-    def serialize(self):
-        result = {0:self.trade, 1:self.quote}        
-        return result, "all"
-    
-    def deserialize(self, s):
-        assert type(s) == list
-        self.trade = s[0][0]
-        self.quote = s[0][1]
-    
-    def find_second_smallest(self, batch, key):
-        smallest = batch[0][key]
-        for i in range(len(batch)):
-            if batch[i][key] > smallest:
-                return batch[i][key]
+        # keys that will never be seen again, safe to delete from the state on the other side
     
     # the execute function signature does not change. stream_id will be a [0 - (length of InputStreams list - 1)] integer
     def execute(self,batches, stream_id, executor_id):
         # state compaction
-        batches = [i for i in batches if len(i) > 0]
+        batches = [i for i in batches if i is not None and len(i) > 0]
         if len(batches) == 0:
             return
-        
-        # self.trade will be a dictionary of lists. 
-        # self.quote will be a dictionary of lists.
+        batch = polars.concat(batches)
+
+        result = None
+        new_left_null = None
 
-        # trade
-        ret_vals = []
         if stream_id == 0:
-            for batch in batches:
-                frames = batch.partition_by(self.group_left_on)
-                for trade_chunk in frames:
-                    symbol = trade_chunk["symbol"][0]
-                    min_trade_ts = trade_chunk[self.left_on][0]
-                    max_trade_ts = trade_chunk[self.left_on][-1]
-                    if symbol not in self.quote:
-                        if symbol in self.trade:
-                            self.trade[symbol].append(trade_chunk)
-                        else:
-                            self.trade[symbol] = [trade_chunk]
-                        continue
-                    current_quotes_for_symbol = self.quote[symbol]
-                    for i in range(len(current_quotes_for_symbol)):
-                        quote_chunk = current_quotes_for_symbol[i]
-                        min_quote_ts = quote_chunk[self.right_on][0]
-                        max_quote_ts = quote_chunk[self.right_on][-1]
-                        #print(max_trade_ts, min_quote_ts, min_trade_ts, max_quote_ts)
-                        if max_trade_ts < min_quote_ts or min_trade_ts > max_quote_ts:
-                            # no overlap.
-                            continue
-                        else:
-                            second_smallest_quote_ts = self.find_second_smallest(quote_chunk, self.right_on)
-                            joinable_trades = trade_chunk[(trade_chunk[self.left_on] >= second_smallest_quote_ts) & (trade_chunk[self.left_on] < max_quote_ts)]
-                            if len(joinable_trades) == 0:
-                                continue
-                            trade_start_ts = joinable_trades[self.left_on][0]
-                            trade_end_ts = joinable_trades[self.left_on][-1]
-                            if len(joinable_trades) == 0:
-                                continue
-                            quote_start_ts = quote_chunk[self.right_on][quote_chunk[self.right_on] <= trade_start_ts][-1]
-                            quote_end_ts = quote_chunk[self.right_on][quote_chunk[self.right_on] <= trade_end_ts][-1]
-                            joinable_quotes = quote_chunk[(quote_chunk[self.right_on] >= quote_start_ts) & (quote_chunk[self.right_on] <= quote_end_ts)]
-                            if len(joinable_quotes) == 0:
-                                continue
-                            trade_chunk = trade_chunk[(trade_chunk[self.left_on] < trade_start_ts) | (trade_chunk[self.left_on] > trade_end_ts)]
-                            new_chunk = quote_chunk[(quote_chunk[self.right_on] < quote_start_ts) | (quote_chunk[self.left_on] > quote_end_ts)]
-                            
-                            self.quote[symbol][i] = new_chunk
-                            
-                            ret_vals.append(joinable_trades.join_asof(joinable_quotes.drop(self.group_right_on), left_on = self.left_on, right_on = self.right_on))
-                            if len(trade_chunk) == 0:
-                                break
-                    
-                    self.quote[symbol] = [i for i in self.quote[symbol] if len(i) > 0]
+            if self.state1 is not None:
+                result = batch.join(self.state1,left_on = self.left_on, right_on = self.right_on ,how=self.batch_how, suffix=self.suffix)
+                if self.how == "left":
+                    new_left_null = batch.join(self.state1, left_on = self.left_on, right_on= self.right_on, how = "anti", suffix = self.suffix)
+            else:
+                if self.how == "left":
+                    new_left_null = batch
 
-                    if len(trade_chunk) == 0:
-                        continue
-                    if symbol in self.trade:
-                        self.trade[symbol].append(trade_chunk)
-                    else:
-                        self.trade[symbol] = [trade_chunk]
-        #quote
+            if self.state0 is None:
+                self.state0 = batch
+            else:
+                self.state0.vstack(batch, in_place = True)
+
+            if self.how == "left" and new_left_null is not None and len(new_left_null) > 0:
+                if self.left_null is None:
+                    self.left_null = new_left_null
+                else:
+                    self.left_null.vstack(new_left_null, in_place= True)
+             
         elif stream_id == 1:
-            for batch in batches:
-                frames = batch.partition_by(self.group_right_on)
-                for quote_chunk in frames:
-                    symbol = quote_chunk["symbol"][0]
-                    min_quote_ts = quote_chunk[self.right_on][0]
-                    max_quote_ts = quote_chunk[self.right_on][-1]
-                    if symbol not in self.trade:
-                        if symbol in self.quote:
-                            self.quote[symbol].append(quote_chunk)
-                        else:
-                            self.quote[symbol] = [quote_chunk]
-                        continue
-                        
-                    current_trades_for_symbol = self.trade[symbol]
-                    for i in range(len(current_trades_for_symbol)):
-                        trade_chunk = current_trades_for_symbol[i]
-                        #print(current_trades_for_symbol)
-                        min_trade_ts = trade_chunk[self.left_on][0]
-                        max_trade_ts = trade_chunk[self.left_on][-1]
-                        if max_trade_ts < min_quote_ts or min_trade_ts > max_quote_ts:
-                            # no overlap.
-                            continue
-                        else:
-                            second_smallest_quote_ts = self.find_second_smallest(quote_chunk, self.right_on)
-                            joinable_trades = trade_chunk[(trade_chunk[self.left_on] >= second_smallest_quote_ts) &( trade_chunk[self.left_on] < max_quote_ts)]
-                            if len(joinable_trades) == 0:
-                                continue
-                            trade_start_ts = joinable_trades[self.left_on][0]
-                            trade_end_ts = joinable_trades[self.left_on][-1]
-                            if len(joinable_trades) == 0:
-                                continue
-                            quote_start_ts = quote_chunk[self.right_on][quote_chunk[self.right_on] <= trade_start_ts][-1]
-                            quote_end_ts = quote_chunk[self.right_on][quote_chunk[self.right_on] <= trade_end_ts][-1]
-                            joinable_quotes = quote_chunk[(quote_chunk[self.right_on] >= quote_start_ts) & (quote_chunk[self.right_on] <= quote_end_ts)]
-                            if len(joinable_quotes) == 0:
-                                continue
-                            quote_chunk = quote_chunk[(quote_chunk[self.right_on] < quote_start_ts ) | (quote_chunk[self.left_on] > quote_end_ts)]
-                            new_chunk = trade_chunk[(trade_chunk[self.left_on] < trade_start_ts) | (trade_chunk[self.left_on] > trade_end_ts)]
-                            
-                            self.trade[symbol][i] = new_chunk
-
-                            ret_vals.append(joinable_trades.join_asof(joinable_quotes.drop(self.group_right_on), left_on = self.left_on, right_on = self.right_on))
-                            if len(quote_chunk) == 0:
-                                break
-                    
-                    self.trade[symbol] = [i for i in self.trade[symbol] if len(i) > 0]
-                    if len(quote_chunk) == 0:
-                        continue
-                    if symbol in self.quote:
-                        self.quote[symbol].append(quote_chunk)
-                    else:
-                        self.quote[symbol] = [quote_chunk]
-        #print(ret_vals)
-
-        if len(ret_vals) == 0:
-            return
-        for thing in ret_vals:
-            print(len(thing))
-            print(thing[thing.symbol=="ZU"])
-        result = polars.concat(ret_vals).drop_nulls()
+            if self.state0 is not None:
+                result = self.state0.join(batch,left_on = self.left_on, right_on = self.right_on ,how=self.batch_how, suffix=self.suffix)
+            
+            if self.how == "left" and self.left_null is not None:
+                self.left_null = self.left_null.join(batch, left_on = self.left_on, right_on = self.right_on, how = "anti", suffix = self.suffix)
+            
+            if self.state1 is None:
+                if self.how == "left":
+                    self.first_row_right = batch[0]
+                self.state1 = batch
+            else:
+                self.state1.vstack(batch, in_place = True)
 
         if result is not None and len(result) > 0:
             return result
     
     def done(self,executor_id):
         #print(len(self.state0),len(self.state1))
-        ret_vals = []
-        for symbol in self.trade:
-            if symbol not in self.quote:
-                continue
-            else:
-                trades = polars.concat(self.trade[symbol]).sort(self.left_on)
-                quotes = polars.concat(self.quote[symbol]).sort(self.right_on)
-                ret_vals.append(trades.join_asof(quotes.drop(self.group_right_on), left_on = self.left_on, right_on = self.right_on, suffix=self.suffix))
-        
-        print("done asof join ", executor_id)
-        return polars.concat(ret_vals).drop_nulls()
+        #print("done join ", executor_id)
+        if self.how == "left" and self.left_null is not None and len(self.left_null) > 0:
+            assert self.first_row_right is not None, "empty RHS"
+            return self.left_null.join(self.first_row_right, left_on= self.left_on, right_on= self.right_on, how = "left", suffix = self.suffix)
 
-class PolarJoinExecutor(Executor):
+
+
+class AntiJoinExecutor(Executor):
     # batch func here expects a list of dfs. This is a quark of the fact that join results could be a list of dfs.
     # batch func must return a list of dfs too
-    def __init__(self, on = None, left_on = None, right_on = None, suffix="_right", how = "inner"):
+    def __init__(self, on = None, left_on = None, right_on = None, suffix="_right"):
 
-        self.state0 = None
+        self.left_null = None
         self.state1 = None
         self.ckpt_start0 = 0
         self.ckpt_start1 = 0
@@ -420,25 +329,10 @@ class PolarJoinExecutor(Executor):
             assert left_on is not None and right_on is not None
             self.left_on = left_on
             self.right_on = right_on
-        self.how = how
+        
+        self.batch_size = 1000000
+        
         # keys that will never be seen again, safe to delete from the state on the other side
-
-    def serialize(self):
-        result = {0:self.state0[self.ckpt_start0:] if (self.state0 is not None and len(self.state0[self.ckpt_start0:]) > 0) else None, 1:self.state1[self.ckpt_start1:] if (self.state1 is not None and len(self.state1[self.ckpt_start1:]) > 0) else None}
-        if self.state0 is not None:
-            self.ckpt_start0 = len(self.state0)
-        if self.state1 is not None:
-            self.ckpt_start1 = len(self.state1)
-        return result, "inc"
-    
-    def deserialize(self, s):
-        assert type(s) == list
-        list0 = [i[0] for i in s if i[0] is not None]
-        list1 = [i[1] for i in s if i[1] is not None]
-        self.state0 = polars.concat(list0) if len(list0) > 0 else None
-        self.state1 = polars.concat(list1) if len(list1) > 0 else None
-        self.ckpt_start0 = len(self.state0) if self.state0 is not None else 0
-        self.ckpt_start1 = len(self.state1) if self.state1 is not None else 0
     
     # the execute function signature does not change. stream_id will be a [0 - (length of InputStreams list - 1)] integer
     def execute(self,batches, stream_id, executor_id):
@@ -448,33 +342,35 @@ class PolarJoinExecutor(Executor):
             return
         batch = polars.concat(batches)
 
-        result = None
+        new_left_null = None
+
         if stream_id == 0:
             if self.state1 is not None:
-                try:
-                    result = batch.join(self.state1,left_on = self.left_on, right_on = self.right_on ,how=self.how, suffix=self.suffix)
-                except:
-                    print(batch)
-            if self.state0 is None:
-                self.state0 = batch
+                new_left_null = batch.join(self.state1, left_on = self.left_on, right_on= self.right_on, how = "anti", suffix = self.suffix)
             else:
-                self.state0.vstack(batch, in_place = True)
+                new_left_null = batch
+
+            if new_left_null is not None and len(new_left_null) > 0:
+                if self.left_null is None:
+                    self.left_null = new_left_null
+                else:
+                    self.left_null.vstack(new_left_null, in_place= True)
              
         elif stream_id == 1:
-            if self.state0 is not None:
-                result = self.state0.join(batch,left_on = self.left_on, right_on = self.right_on ,how=self.how, suffix=self.suffix)
+            if self.left_null is not None:
+                self.left_null = self.left_null.join(batch, left_on = self.left_on, right_on = self.right_on, how = "anti", suffix = self.suffix)
+            
             if self.state1 is None:
                 self.state1 = batch
             else:
                 self.state1.vstack(batch, in_place = True)
-
-        if result is not None and len(result) > 0:
-            return result
     
     def done(self,executor_id):
         #print(len(self.state0),len(self.state1))
         #print("done join ", executor_id)
-        pass
+        if self.left_null is not None and len(self.left_null) > 0:
+            for i in range(0, len(self.left_null), self.batch_size):
+                yield self.left_null[i: i + self.batch_size]
 
 class DistinctExecutor(Executor):
     def __init__(self, keys) -> None:
