@@ -9,7 +9,7 @@ import sys
 import polars
 import pyarrow as pa
 import types
-import duckdb
+#import duckdb
 import psutil
 import pyarrow.flight
 # isolated simplified test bench for different fault tolerance protocols
